#include <QDebugStateSaver>

#include "triangle.h"

Triangle::Shared_Edge::Edge rotate_edge_enum(Triangle::Shared_Edge::Edge edge);

Triangle::Triangle() {}

Triangle::Triangle(const Triangle &other) {
    this->vertex_buffer_ = other.vertex_buffer_;
    this->u_ = other.u_;
    this->v_ = other.v_;
    this->w_ = other.w_;
}

Triangle::Triangle(const QVector<Vertex> &vertex_buffer, int u, int v, int w) {
    this->vertex_buffer_ = vertex_buffer;
    this->u_ = u;
    this->v_ = v;
    this->w_ = w;
}

Vertex Triangle::u() const {
    return vertex_buffer_[u_];
}

Vertex Triangle::v() const {
    return vertex_buffer_[v_];
}

Vertex Triangle::w() const {
    return vertex_buffer_[w_];
}

int Triangle::u_idx() const {
    return u_;
}

int Triangle::v_idx() const {
    return v_;
}

int Triangle::w_idx() const {
    return w_;
}

bool Triangle::get_shared_edge(Triangle other, Shared_Edge &edge) const {
    if (get_shared_edge_(other, edge)) {
        return true;
    }

    other.rotate_indices();
    if (get_shared_edge_(other, edge)) {
        edge.edge_b = rotate_edge_enum(rotate_edge_enum(edge.edge_b));
        return true;
    }

    other.rotate_indices();
    if (get_shared_edge_(other, edge)) {
        edge.edge_b = rotate_edge_enum(edge.edge_b);
        return true;
    }

    return false;
}

bool Triangle::get_shared_edge_(const Triangle &other, Shared_Edge &edge) const {
    if (u().samePos(other.u())) {
        if (v().samePos(other.w())) {
            edge.edge_a = Triangle::Shared_Edge::Edge::uv;
            edge.edge_b = Triangle::Shared_Edge::Edge::wu;

            edge.a1 = u_idx();
            edge.b1 = other.w_idx();

            edge.a2 = v_idx();
            edge.b2 = other.u_idx();

            edge.a3 = w_idx();
            edge.b3 = other.v_idx();

            return true;
        } else if (w().samePos(other.v())) {
            edge.edge_a = Triangle::Shared_Edge::Edge::wu;
            edge.edge_b = Triangle::Shared_Edge::Edge::uv;

            edge.a1 = u_idx();
            edge.b1 = other.u_idx();

            edge.a2 = w_idx();
            edge.b2 = other.v_idx();

            edge.a3 = v_idx();
            edge.b3 = other.w_idx();

            return true;
        }
    } else if (v().samePos(other.u()) && w().samePos(other.w())) {
        edge.edge_a = Triangle::Shared_Edge::Edge::vw;
        edge.edge_b = Triangle::Shared_Edge::Edge::wu;

        edge.a1 = v_idx();
        edge.b1 = other.w_idx();

        edge.a2 = w_idx();
        edge.b2 = other.u_idx();

        edge.a3 = u_idx();
        edge.b3 = other.v_idx();

        return true;
    }

    return false;
}

Triangle &Triangle::operator=(const Triangle &other) {
    this->vertex_buffer_ = other.vertex_buffer_;
    this->u_ = other.u_;
    this->v_ = other.v_;
    this->w_ = other.w_;
    return *this;
}

bool Triangle::operator==(const Triangle &other) const {
    return u_ == other.u_ && v_ == other.v_ && w_ == other.w_;
}

bool Triangle::operator<(const Triangle &other) const {
    if (u_ < other.u_) {
        return true;
    } else if (u_ > other.u_) {
        return false;
    }

    if (v_ < other.v_) {
        return true;
    } else if (v_ > other.v_) {
        return false;
    }

    if (w_ < other.w_) {
        return true;
    } else if (w_ > other.w_) {
        return false;
    }

    return false;
}

void Triangle::rotate_indices() {
    unsigned int a = u_;
    u_ = v_;
    v_ = w_;
    w_ = a;
}

Triangle::Shared_Edge::Edge rotate_edge_enum(Triangle::Shared_Edge::Edge edge) {
    switch (edge) {
        case Triangle::Shared_Edge::Edge::uv:
            return Triangle::Shared_Edge::Edge::vw;
        case Triangle::Shared_Edge::Edge::vw:
            return Triangle::Shared_Edge::Edge::wu;
        case Triangle::Shared_Edge::Edge::wu:
            return Triangle::Shared_Edge::Edge::uv;
        default:
            qDebug() << "Default case should be unreachable";
            return Triangle::Shared_Edge::Edge::uv;
    }
}

QDebug operator<<(QDebug d, const Triangle &triangle) {
    QDebugStateSaver saver(d);
    d.nospace() << "Triangle(" << triangle.u_idx() << ", " << triangle.v_idx() << ", " << triangle.w_idx() << "; " << triangle.u().pos << ", " << triangle.v().pos << ", " << triangle.w().pos << ")";
    return d;
}
