#pragma once
#ifndef TRIANGLE_H
#define TRIANGLE_H

#include <QDebug>
#include <QVector>
#include "vertex.h"

class Triangle {
    public:
        /**
         * Stores the indices of the shared edge of two triangles as well as the other two indices.
         *
         * a_* are the indices of the first triangle, b_* those of the second triangle.
         *
         * The vertices at a_1 and b_1 are at the same position as are those at a_2 and b_2.
         *
         * edge_a describes which edge of the first triangle is part of the shared edge.
         */
        struct Shared_Edge {
            enum Edge {
                uv,
                vw,
                wu
            };

            Edge edge_a;
            Edge edge_b;
            unsigned int a1;
            unsigned int b1;
            unsigned int a2;
            unsigned int b2;
            unsigned int a3;
            unsigned int b3;
        };

        Triangle();
        Triangle(const Triangle &other);
        Triangle(const QVector<Vertex> &vertex_buffer, int u, int v, int w);

        Vertex u() const;
        Vertex v() const;
        Vertex w() const;

        int u_idx() const;
        int v_idx() const;
        int w_idx() const;

        bool get_shared_edge(Triangle other, Shared_Edge &edge) const;

        Triangle &operator=(const Triangle &other);

        // == and < both ignore the vertex buffer, they only compare indices.
        // They are implemented so that Triangle can be used as the key of a QMap.
        bool operator==(const Triangle &other) const;
        bool operator<(const Triangle &other) const;

    private:
        QVector<Vertex> vertex_buffer_;
        int u_;
        int v_;
        int w_;

        void rotate_indices();
        bool get_shared_edge_(const Triangle &other, Shared_Edge &edge) const;
};

QDebug operator<<(QDebug d, const Triangle &triangle);

#endif
